class MetricsAgent:
    """
    Any metric client should implement this interface
    to be compatible with subscriber.subscribe
    """
    def histogram(self,
                  metric: str,
                  value: float) -> None:
        pass

    def increment(self,
                  metric: str,
                  value: float = 1) -> None:
        pass



from gcloud.aio.auth import BUILD_GCLOUD_REST

if BUILD_GCLOUD_REST:
    pass
else:
    import aiohttp
    import asyncio
    import logging
    import time
    from typing import Awaitable
    from typing import Callable
    from typing import List
    from typing import Optional
    from typing import Tuple
    from typing import TYPE_CHECKING
    from typing import TypeVar

    from gcloud.aio.pubsub.subscriber_client import SubscriberClient
    from gcloud.aio.pubsub.subscriber_message import SubscriberMessage
    from gcloud.aio.pubsub.metrics_agent import MetricsAgent

    log = logging.getLogger(__name__)

    if TYPE_CHECKING:
        MessageQueue = asyncio.Queue[Tuple[SubscriberMessage,  # pylint: disable=unsubscriptable-object
                                           float]]
    else:
        MessageQueue = asyncio.Queue
    ApplicationHandler = Callable[[SubscriberMessage], Awaitable[None]]
    T = TypeVar('T')

    class AckDeadlineCache:
        def __init__(self, subscriber_client: SubscriberClient,
                     subscription: str, cache_timeout: float):
            self.subscriber_client = subscriber_client
            self.subscription = subscription
            self.cache_timeout = cache_timeout
            self.ack_deadline: float = float('inf')
            self.last_refresh: float = float('-inf')

        async def get(self) -> float:
            if self.cache_outdated():
                await self.refresh()
            return self.ack_deadline

        async def refresh(self) -> None:
            try:
                sub = await self.subscriber_client.get_subscription(
                    self.subscription)
                self.ack_deadline = float(sub['ackDeadlineSeconds'])
            except Exception as e:
                log.warning(
                    'Failed to refresh ackDeadlineSeconds value', exc_info=e)
            self.last_refresh = time.perf_counter()

        def cache_outdated(self) -> bool:
            if (time.perf_counter() - self.last_refresh > self.cache_timeout
                    or self.ack_deadline == float('inf')):
                return True
            return False

    async def _budgeted_queue_get(queue: 'asyncio.Queue[T]',
                                  time_budget: float) -> List[T]:
        result = []
        while time_budget > 0:
            start = time.perf_counter()
            try:
                message = await asyncio.wait_for(
                    queue.get(), timeout=time_budget)
                result.append(message)
            except asyncio.TimeoutError:
                break
            time_budget -= (time.perf_counter() - start)
        return result

    async def acker(subscription: str,
                    ack_queue: 'asyncio.Queue[str]',
                    subscriber_client: 'SubscriberClient',
                    ack_window: float,
                    metrics_client: MetricsAgent) -> None:
        ack_ids: List[str] = []
        while True:
            if not ack_ids:
                ack_ids.append(await ack_queue.get())

            ack_ids += await _budgeted_queue_get(ack_queue, ack_window)

            # acknowledge endpoint limit is 524288 bytes
            # which is ~2744 ack_ids
            if len(ack_ids) > 2500:
                log.error(
                    'acker is falling behind, dropping %d unacked messages',
                    len(ack_ids) - 2500)
                ack_ids = ack_ids[-2500:]
                for _ in range(len(ack_ids) - 2500):
                    ack_queue.task_done()

            try:
                await subscriber_client.acknowledge(subscription,
                                                    ack_ids=ack_ids)
                for _ in ack_ids:
                    ack_queue.task_done()
            except aiohttp.client_exceptions.ClientResponseError as e:
                if e.status == 400:
                    log.error(
                        'Ack error is unrecoverable, '
                        'one or more messages may be dropped', exc_info=e)

                    async def maybe_ack(ack_id: str) -> None:
                        try:
                            await subscriber_client.acknowledge(
                                subscription,
                                ack_ids=[ack_id])
                        except Exception as e:
                            log.warning('Ack failed for ack_id=%s',
                                        ack_id,
                                        exc_info=e)
                        finally:
                            ack_queue.task_done()

                    for ack_id in ack_ids:
                        asyncio.ensure_future(maybe_ack(ack_id))
                    ack_ids = []

                log.warning(
                    'Ack request failed, better luck next batch', exc_info=e)
                metrics_client.increment('pubsub.acker.batch.failed')

                continue
            except asyncio.CancelledError:  # pylint: disable=try-except-raise
                raise
            except Exception as e:
                log.warning(
                    'Ack request failed, better luck next batch', exc_info=e)
                metrics_client.increment('pubsub.acker.batch.failed')

                continue

            metrics_client.histogram('pubsub.acker.batch', len(ack_ids))

            ack_ids = []

    async def nacker(subscription: str,
                     nack_queue: 'asyncio.Queue[str]',
                     subscriber_client: 'SubscriberClient',
                     nack_window: float,
                     metrics_client: MetricsAgent) -> None:
        ack_ids: List[str] = []
        while True:
            if not ack_ids:
                ack_ids.append(await nack_queue.get())

            ack_ids += await _budgeted_queue_get(nack_queue, nack_window)

            # modifyAckDeadline endpoint limit is 524288 bytes
            # which is ~2744 ack_ids
            if len(ack_ids) > 2500:
                log.error(
                    'nacker is falling behind, dropping %d unacked messages',
                    len(ack_ids) - 2500)
                ack_ids = ack_ids[-2500:]
                for _ in range(len(ack_ids) - 2500):
                    nack_queue.task_done()
            try:
                await subscriber_client.modify_ack_deadline(
                    subscription,
                    ack_ids=ack_ids,
                    ack_deadline_seconds=0)
                for _ in ack_ids:
                    nack_queue.task_done()
            except aiohttp.client_exceptions.ClientResponseError as e:
                if e.status == 400:
                    log.error(
                        'Nack error is unrecoverable, '
                        'one or more messages may be dropped', exc_info=e)

                    async def maybe_nack(ack_id: str) -> None:
                        try:
                            await subscriber_client.modify_ack_deadline(
                                subscription,
                                ack_ids=[ack_id],
                                ack_deadline_seconds=0)
                        except Exception as e:
                            log.warning('Nack failed for ack_id=%s',
                                        ack_id,
                                        exc_info=e)
                        finally:
                            nack_queue.task_done()
                    for ack_id in ack_ids:
                        asyncio.ensure_future(maybe_nack(ack_id))
                    ack_ids = []

                log.warning(
                    'Nack request failed, better luck next batch', exc_info=e)
                metrics_client.increment('pubsub.nacker.batch.failed')

                continue
            except asyncio.CancelledError:  # pylint: disable=try-except-raise
                raise
            except Exception as e:
                log.warning(
                    'Nack request failed, better luck next batch', exc_info=e)
                metrics_client.increment('pubsub.nacker.batch.failed')

                continue

            metrics_client.histogram('pubsub.nacker.batch', len(ack_ids))

            ack_ids = []

    async def _execute_callback(message: SubscriberMessage,
                                callback: ApplicationHandler,
                                ack_queue: 'asyncio.Queue[str]',
                                nack_queue: 'Optional[asyncio.Queue[str]]',
                                metrics_client: MetricsAgent
                                ) -> None:
        try:
            start = time.perf_counter()
            await callback(message)
            await ack_queue.put(message.ack_id)
            metrics_client.increment('pubsub.consumer.succeeded')
            metrics_client.histogram('pubsub.consumer.latency.runtime',
                                     time.perf_counter() - start)
        except asyncio.CancelledError:
            if nack_queue:
                await nack_queue.put(message.ack_id)
            log.warning('Application callback was cancelled')
            metrics_client.increment('pubsub.consumer.cancelled')
        except Exception:
            if nack_queue:
                await nack_queue.put(message.ack_id)
            log.exception('Application callback raised an exception')
            metrics_client.increment('pubsub.consumer.failed')

    async def consumer(  # pylint: disable=too-many-locals
            message_queue: MessageQueue,
            callback: ApplicationHandler,
            ack_queue: 'asyncio.Queue[str]',
            ack_deadline_cache: AckDeadlineCache,
            max_tasks: int,
            nack_queue: 'Optional[asyncio.Queue[str]]',
            metrics_client: MetricsAgent) -> None:
        try:
            semaphore = asyncio.Semaphore(max_tasks)

            async def _consume_one(message: SubscriberMessage,
                                   pulled_at: float) -> None:
                await semaphore.acquire()

                ack_deadline = await ack_deadline_cache.get()
                if (time.perf_counter() - pulled_at) >= ack_deadline:
                    metrics_client.increment('pubsub.consumer.failfast')
                    message_queue.task_done()
                    semaphore.release()
                    return

                metrics_client.histogram(
                    'pubsub.consumer.latency.receive',
                    # publish_time is in UTC Zulu
                    # https://cloud.google.com/pubsub/docs/reference/rest/v1/PubsubMessage
                    time.time() - message.publish_time.timestamp())

                task = asyncio.ensure_future(_execute_callback(
                    message,
                    callback,
                    ack_queue,
                    nack_queue,
                    metrics_client,
                ))
                task.add_done_callback(lambda _f: semaphore.release())
                message_queue.task_done()

            while True:
                message, pulled_at = await message_queue.get()
                await asyncio.shield(_consume_one(message, pulled_at))
        except asyncio.CancelledError:
            log.info('Consumer worker cancelled. Gracefully terminating...')
            for _ in range(max_tasks):
                await semaphore.acquire()

            await ack_queue.join()
            if nack_queue:
                await nack_queue.join()
            log.info('Consumer terminated gracefully.')
            raise

    async def producer(
            subscription: str,
            message_queue: MessageQueue,
            subscriber_client: 'SubscriberClient',
            max_messages: int,
            metrics_client: MetricsAgent) -> None:
        try:
            while True:
                new_messages = []
                try:
                    pull_task = asyncio.ensure_future(
                        subscriber_client.pull(
                            subscription=subscription,
                            max_messages=max_messages,
                            # it is important to have this value reasonably
                            # high as long lived connections may be left
                            # hanging on a server which will cause delay in
                            # message delivery or even false deadlettering if
                            # it is enabled
                            timeout=30))
                    new_messages = await asyncio.shield(pull_task)
                except (asyncio.TimeoutError, KeyError):
                    continue

                metrics_client.histogram(
                    'pubsub.producer.batch', len(new_messages))

                pulled_at = time.perf_counter()
                while new_messages:
                    await message_queue.put((new_messages[-1], pulled_at))
                    new_messages.pop()

                await message_queue.join()
        except asyncio.CancelledError:
            log.info('Producer worker cancelled. Gracefully terminating...')

            if not pull_task.done():
                # Leaving the connection hanging can result in redelivered
                # messages, so try to finish before shutting down
                try:
                    new_messages += await asyncio.wait_for(pull_task, 5)
                except (asyncio.TimeoutError, KeyError):
                    pass

            pulled_at = time.perf_counter()
            for m in new_messages:
                await message_queue.put((m, pulled_at))

            await message_queue.join()
            log.info('Producer terminated gracefully.')
            raise

    async def subscribe(subscription: str,  # pylint: disable=too-many-locals
                        handler: ApplicationHandler,
                        subscriber_client: SubscriberClient,
                        *,
                        num_producers: int = 1,
                        max_messages_per_producer: int = 100,
                        ack_window: float = 0.3,
                        ack_deadline_cache_timeout: float = float('inf'),
                        num_tasks_per_consumer: int = 1,
                        enable_nack: bool = True,
                        nack_window: float = 0.3,
                        metrics_client: Optional[MetricsAgent] = None
                        ) -> None:
        ack_queue: 'asyncio.Queue[str]' = asyncio.Queue(
            maxsize=(max_messages_per_producer * num_producers))
        nack_queue: 'Optional[asyncio.Queue[str]]' = None
        ack_deadline_cache = AckDeadlineCache(subscriber_client,
                                              subscription,
                                              ack_deadline_cache_timeout)
        metrics_client = metrics_client or MetricsAgent()
        acker_tasks = []
        consumer_tasks = []
        producer_tasks = []
        try:
            acker_tasks.append(asyncio.ensure_future(
                acker(subscription, ack_queue, subscriber_client,
                      ack_window=ack_window, metrics_client=metrics_client)
            ))
            if enable_nack:
                nack_queue = asyncio.Queue(
                    maxsize=(max_messages_per_producer * num_producers))
                acker_tasks.append(asyncio.ensure_future(
                    nacker(subscription, nack_queue, subscriber_client,
                           nack_window=nack_window,
                           metrics_client=metrics_client)
                ))
            for _ in range(num_producers):
                q: MessageQueue = asyncio.Queue(
                    maxsize=max_messages_per_producer)
                consumer_tasks.append(asyncio.ensure_future(
                    consumer(q,
                             handler,
                             ack_queue,
                             ack_deadline_cache,
                             num_tasks_per_consumer,
                             nack_queue,
                             metrics_client=metrics_client)
                ))
                producer_tasks.append(asyncio.ensure_future(
                    producer(subscription,
                             q,
                             subscriber_client,
                             max_messages=max_messages_per_producer,
                             metrics_client=metrics_client)
                ))

            # TODO: since this is in a `not BUILD_GCLOUD_REST` section, we
            # shouldn't have to care about py2 support. Using splat syntax
            # here, though, breaks the coverage.py reporter for this file even
            # though it would never be loaded at runtime in py2.
            # all_tasks = [*producer_tasks, *consumer_tasks, *acker_tasks]
            all_tasks = producer_tasks + consumer_tasks + acker_tasks
            done, _ = await asyncio.wait(all_tasks,
                                         return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                task.result()
            raise Exception('A subscriber worker shut down unexpectedly!')
        except (asyncio.CancelledError, Exception) as e:
            log.info('Subscriber exited', exc_info=e)
            for task in producer_tasks:
                task.cancel()
            await asyncio.wait(producer_tasks,
                               return_when=asyncio.ALL_COMPLETED)

            for task in consumer_tasks:
                task.cancel()
            await asyncio.wait(consumer_tasks,
                               return_when=asyncio.ALL_COMPLETED)

            for task in acker_tasks:
                task.cancel()
            await asyncio.wait(acker_tasks, return_when=asyncio.ALL_COMPLETED)
        raise asyncio.CancelledError('Subscriber shut down')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcloud', 'gcloud.aio', 'gcloud.aio.pubsub']

package_data = \
{'': ['*']}

install_requires = \
['gcloud-aio-auth>=3.3.0,<5.0.0']

setup_kwargs = {
    'name': 'gcloud-aio-pubsub',
    'version': '5.0.1',
    'description': 'Python Client for Google Cloud Pub/Sub',
    'long_description': "(Asyncio OR Threadsafe) Python Client for Google Cloud Pub/Sub\n==============================================================\n\n    This is a shared codebase for ``gcloud-aio-pubsub`` and\n    ``gcloud-rest-pubsub``\n\n|pypi| |pythons-aio| |pythons-rest|\n\nInstallation\n------------\n\n.. code-block:: console\n\n    $ pip install --upgrade gcloud-{aio,rest}-pubsub\n\nUsage\n-----\n\nSubscriber\n~~~~~~~~~~\n\n``gcloud-{aio,rest}-pubsub`` provides ``SubscriberClient``\nas an interface to call pubsub's HTTP API:\n\n.. code-block:: python\n\n    from gcloud.aio.pubsub import SubscriberClient\n    from gcloud.aio.pubsub import SubscriberMessage\n\n    client = SubscriberClient()\n    # create subscription\n    await client.create_subscription(\n        'projects/<project_name>/subscriptions/<subscription_name>',\n        'projects/<project_name>/topics/<topic_name>')\n\n    # pull messages\n    messages: List[SubscriberMessage] = await client.pull(\n        'projects/<project_name>/subscriptions/<subscription_name>',\n        max_messages=10)\n\n\nThere's also ``gcloud.aio.pubsub.subscribe`` helper function you can use to\nsetup a pubsub processing pipeline. It is built with ``asyncio`` and thus only\navailable in ``gcloud-aio-pubsub`` package. The usage is fairly simple:\n\n.. code-block:: python\n\n    from gcloud.aio.pubsub import SubscriberClient\n    from gcloud.aio.pubsub import subscribe\n    from gcloud.aio.pubsub.metrics_agent import MetricsAgent\n\n    subscriber_client = SubscriberClient()\n\n    async def handler(message):\n        return\n\n    await subscribe(\n        'projects/<my_project>/subscriptions/<my_subscription>',\n        handler,\n        subscriber_client,\n        num_producers=1,\n        max_messages_per_producer=100,\n        ack_window=0.3,\n        num_tasks_per_consumer=1,\n        enable_nack=True,\n        nack_window=0.3,\n        metrics_client=MetricsAgent()\n    )\n\nWhile defaults are somewhat sensible, it is highly recommended to performance\ntest your application and tweak function parameter to your specific needs.\nHere's a few hints:\n\n:``handler``:\n    an async function that will be called for each message. It should accept an\n    instance of ``SubscriberMessage`` as its only argument and return ``None``\n    if the message should be acked. An exception raised within the handler will\n    result in the message being left to expire, and thus it will be redelivered\n    according to your subscription's ack deadline.\n\n:``num_producers``:\n    number of workers that will be making ``pull`` requests to pubsub. Please\n    note that a worker will only fetch new batch once the ``handler`` was called\n    for each message from the previous batch. This means that running only a\n    single worker will most likely make your application IO bound. If you notice\n    this being an issue don't hesitate to bump this parameter.\n\n:``max_messages_per_producer``:\n    number of pubsub messages a worker will try to fetch in a single batch. This\n    value is passed to ``pull`` `endpoint`_ as ``maxMessages`` parameter. A rule\n    of thumb here is the faster your handler is the bigger this value should be.\n\n:``ack_window``:\n    ack requests are handled separately and are done in batches. This parameters\n    specifies how often ack requests will be made. Setting it to ``0.0`` will\n    effectively disable batching.\n\n:``num_tasks_per_consumer``:\n    how many ``handle`` calls a worker can make until it blocks to wait for them\n    to return. If you process messages independently from each other you should\n    be good with the default value of ``1``. If you do something fancy (e.g.\n    aggregate messages before processing them), you'll want a higher pool here.\n    You can think of ``num_producers * num_tasks_per_consumer`` as an upper\n    limit of how many messages can possibly be within your application state at\n    any given moment.\n\n:``enable_nack``:\n    if enabled messages for which ``callback`` raised an exception will be\n    explicitly nacked using ``modifyAckDeadline`` endpoint so they can be\n    retried immediately.\n\n:``nack_window``:\n    same as ``ack_window`` but for nack requests\n\n\n``subscribe`` has also an optional ``metrics_client`` argument. You can provide\nany metrics agent that implements the same interface as ``MetricsAgent``\n(Datadog client will do ;) ) and get the following metrics:\n\n- ``pubsub.producer.batch`` - [histogram] actual size of a batch retrieved from\n  pubsub.\n\n- ``pubsub.consumer.failfast`` - [increment] a message was dropped due to its\n  lease being expired.\n\n- ``pubsub.consumer.latency.receive`` - [histogram] how many seconds it took for\n  a message to reach handler after it was published.\n\n- ``pubsub.consumer.succeeded`` - [increment] ``handler`` call was successfull.\n\n- ``pubsub.consumer.failed`` - [increment] ``handler`` call raised an exception.\n\n- ``pubsub.consumer.latency.runtime`` - [histogram] ``handler`` execution time\n  in seconds.\n\n- ``pubsub.acker.batch.failed`` - [increment] ack request failed.\n\n- ``pubsub.acker.batch`` - [histogram] actual number of messages that was acked\n  in a single request.\n\n\nPublisher\n~~~~~~~~~\n\nThe ``PublisherClient`` is a dead-simple alternative to the official Google\nCloud Pub/Sub publisher client. The main design goal was to eliminate all the\nadditional gRPC overhead implemented by the upstream client.\n\nIf migrating between this library and the official one, the main difference is\nthis: the ``gcloud-{aio,rest}-pubsub`` publisher's ``.publish()`` method *immediately*\npublishes the messages you've provided, rather than maintaining our own\npublishing queue, implementing batching and flow control, etc. If you're\nlooking for a full-featured publishing library with all the bells and whistles\nbuilt in, you may be interested in the upstream provider. If you're looking to\nmanage your own batching / timeouts / retry / threads / etc, this library\nshould be a bit easier to work with.\n\nSample usage:\n\n.. code-block:: python\n\n    from gcloud.aio.pubsub import PubsubMessage\n    from gcloud.aio.pubsub import PublisherClient\n\n    async with aiohttp.ClientSession() as session:\n        client = PublisherClient(session=session)\n\n        topic = client.topic_path('my-gcp-project', 'my-topic-name')\n\n        messages = [\n            PubsubMessage(b'payload', attribute='value'),\n            PubsubMessage(b'other payload', other_attribute='whatever',\n                          more_attributes='something else'),\n        ]\n        response = await client.publish(topic, messages)\n        # response == {'messageIds': ['1', '2']}\n\nEmulators\n^^^^^^^^^\n\nFor testing purposes, you may want to use ``gcloud-aio-pubsub`` along with a\nlocal GCS emulator. Setting the ``$PUBSUB_EMULATOR_HOST`` environment variable\nto the local address of your emulator should be enough to do the trick.\n\nFor example, using the official Google Pubsub emulator:\n\n.. code-block:: console\n\n    gcloud beta emulators pubsub start --host-port=0.0.0.0:8681\n    export PUBSUB_EMULATOR_HOST='0.0.0.0:8681'\n\nAny ``gcloud-aio-pubsub`` Publisher requests made with that environment\nvariable set will query the emulator instead of the official GCS APIs.\n\nFor easier ergonomics, you may be interested in\n`messagebird/gcloud-pubsub-emulator`_.\n\nCustomization\n-------------\n\nThis library mostly tries to stay agnostic of potential use-cases; as such, we\ndo not implement any sort of retrying or other policies under the assumption\nthat we wouldn't get things right for every user's situation.\n\nAs such, we recommend configuring your own policies on an as-needed basis. The\n`backoff`_ library can make this quite straightforward! For example, you may\nfind it useful to configure something like:\n\n.. code-block:: python\n\n    class SubscriberClientWithBackoff(SubscriberClient):\n        @backoff.on_exception(backoff.expo, aiohttp.ClientResponseError,\n                              max_tries=5, jitter=backoff.full_jitter)\n        async def pull(self, *args: Any, **kwargs: Any):\n            return await super().pull(*args, **kwargs)\n\nContributing\n------------\n\nPlease see our `contributing guide`_.\n\n.. _contributing guide: https://github.com/talkiq/gcloud-aio/blob/master/.github/CONTRIBUTING.rst\n.. _messagebird/gcloud-pubsub-emulator: https://github.com/marcelcorso/gcloud-pubsub-emulator#gcloud-pubsub-emulator\n.. _official Google documentation: https://github.com/googleapis/google-cloud-python/blob/11c72ade8b282ae1917fba19e7f4e0fe7176d12b/pubsub/google/cloud/pubsub_v1/gapic/subscriber_client.py#L236\n.. _backoff: https://pypi.org/project/backoff/\n\n.. |pypi| image:: https://img.shields.io/pypi/v/gcloud-aio-pubsub.svg?style=flat-square\n    :alt: Latest PyPI Version\n    :target: https://pypi.org/project/gcloud-aio-pubsub/\n\n.. |pythons-aio| image:: https://img.shields.io/pypi/pyversions/gcloud-aio-pubsub.svg?style=flat-square&label=python (aio)\n    :alt: Python Version Support (gcloud-aio-pubsub)\n    :target: https://pypi.org/project/gcloud-aio-pubsub/\n\n.. |pythons-rest| image:: https://img.shields.io/pypi/pyversions/gcloud-rest-pubsub.svg?style=flat-square&label=python (rest)\n    :alt: Python Version Support (gcloud-rest-pubsub)\n    :target: https://pypi.org/project/gcloud-rest-pubsub/\n\n.. _endpoint: https://cloud.google.com/pubsub/docs/reference/rest/v1/projects.subscriptions/pull#request-body\n",
    'author': 'Vi Engineering',
    'author_email': 'voiceai-eng@dialpad.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/talkiq/gcloud-aio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
